#!/bin/bash

# -------------------------------------------------------
# Globals
# -------------------------------------------------------
BIN_NAME="enable_vrx"

FIRMWARE_PATH="/firmware/image/qcn9000/"
BDF_FILE_NAME="bdwlan.bac"
BDF_FILE_NAME_2_4_G="bdwlan.bac_minipini2G"
BDF_FILE_NAME_5_G="bdwlan.bac_minipini5G"

# -------------------------------------------------------
# Array of valid 2.4GHz channels
# -------------------------------------------------------
declare -A channel_array_2_4g=(
	[1]="1 - 2412 - 20"
	[2]="2 - 2417 - 20"
	[3]="3 - 2422 - 20"
	[4]="4 - 2427 - 20"
	[5]="5 - 2432 - 20"
	[6]="6 - 2437 - 20"
	[7]="7 - 2442 - 20"
	[8]="8 - 2447 - 20"
	[9]="9 - 2452 - 20"
	[10]="10 - 2457 - 20"
	[11]="11 - 2462 - 20"
	[12]="12 - 2467 - 20"
	[13]="13 - 2472 - 20"
)

# -------------------------------------------------------
# Array of valid 5GHz channels
# -------------------------------------------------------
declare -A channel_array_5g=(
	[36]="36 - 5180 - 20"
	[40]="40 - 5200 - 20"
	[44]="44 - 5220 - 20"
	[48]="48 - 5240 - 20"
	[52]="52 - 5260 - 20"
	[56]="56 - 5280 - 20"
	[60]="60 - 5300 - 20"
	[64]="64 - 5320 - 20"
	[100]="100 - 5500 - 20"
	[104]="104 - 5520 - 20"
	[108]="108 - 5540 - 20"
	[112]="112 - 5560 - 20"
	[116]="116 - 5580 - 20"
	[120]="120 - 5600 - 20"
	[124]="124 - 5620 - 20"
	[128]="128 - 5640 - 20"
	[132]="132 - 5660 - 20"
	[136]="136 - 5680 - 20"
	[140]="140 - 5700 - 20"
	[144]="144 - 5720 - 20"
	[149]="149 - 5745 - 20"
	[153]="153 - 5765 - 20"
	[157]="157 - 5785 - 20"
	[161]="161 - 5805 - 20"
	[165]="165 - 5825 - 20"
	[173]="173 - 5865 - 20"
	[177]="177 - 5885 - 20"
)

# -------------------------------------------------------
# Metadata for each band
# FREQ_BAND: Frequency band (i.e. 2.4GHz / 5GHz)
# FREQ_BAND_ID: ieee80211 WLAN band ID. Used to configure channel 
# FREQ_BAND_MODE: Mode used for the frequency band
# FREQ_BAND_CHANNEL: User provided channe;
# FREQ_BAND_FIRMWARE_NAME: bdwlan.bac firmare bin for the band
# FREQ_BAND_CHANNEL_ARRAY: list of accepted channels for the band
# -------------------------------------------------------
FREQ_BAND=""
FREQ_BAND_ID=""
FREQ_BAND_MODE=""
FREQ_BAND_CHANNEL=""
FREQ_BAND_FIRMWARE_NAME=""
declare -A FREQ_BAND_CHANNEL_ARRAY

# -------------------------------------------------------
# Utility functions
# -------------------------------------------------------
print_usage() {
    local HEADER_COLOR='\033[1;32m'  # Green
    local RESET_COLOR='\033[0m'      # Reset color

    # Print the description first
    echo -e "${HEADER_COLOR}Description:${RESET_COLOR}"
    echo -e "  Utility script used by voxl vrx to set the ath0 interface into monitor mode."
    echo -e "  You can specify the frequency band (2.4GHz or 5GHz) and select a specific channel"
    echo ""

    # Print the usage
    echo -e "${HEADER_COLOR}Usage:${RESET_COLOR}"
    echo -e "  $BIN_NAME -c <channel_number> - Set channel"
    echo -e "  $BIN_NAME -b <band>           - Set frequency band (2G or 5G)"
    echo -e "  $BIN_NAME -h                  - Show this help menu"
    echo ""
    
    # Print example
    echo -e "${HEADER_COLOR}Example:${RESET_COLOR}"
    echo -e "  $BIN_NAME -b 5G -c 161         - Set 5GHz band with channel 161"
    echo -e "  $BIN_NAME -b 2G -c 11        - Set 2.4GHz band with channel 11"
    echo ""
}

print_channels() {
    local HEADER_COLOR='\033[1;32m'   # Blue
    local TABLE_COLOR='\033[1;34m'    # Green
    local RESET_COLOR='\033[0m'       # Reset color

    echo -e "${TABLE_COLOR}Table for $FREQ_BAND Band${RESET_COLOR}"
    echo "----------------------------------------"
    printf "${HEADER_COLOR}%-8s %-10s %-10s${RESET_COLOR}\n" "Channel" "Frequency" "Bandwidth"
    echo "----------------------------------------"
    
    for channel in $(echo "${!FREQ_BAND_CHANNEL_ARRAY[@]}" | tr ' ' '\n' | sort -n); do
        IFS=" - " read -r chan freq bw <<< "${FREQ_BAND_CHANNEL_ARRAY[$channel]}"
        
        printf "%-8s %-10s %-10s\n" "$chan" "$freq" "$bw"
    done

    echo "----------------------------------------"
	echo ""
}

print_info() {
    echo "$BIN_NAME: INFO: $1"
}

print_error() {
    echo "$BIN_NAME: ERROR: $1"
}

# -------------------------------------------------------
# Configure metadata based on selected band
# -------------------------------------------------------
setup_freq_band() {
	declare -n channel_array_name

    case "$1" in
        "2G")
			FREQ_BAND=$1
			FREQ_BAND_ID=1
			FREQ_BAND_MODE="11NGHT20"
			FREQ_BAND_FIRMWARE_NAME=$BDF_FILE_NAME_2_4_G
			channel_array_name="channel_array_2_4g"
            ;;
        "5G")
			FREQ_BAND=$1
			FREQ_BAND_ID=2
			FREQ_BAND_MODE="11AHE20"
			FREQ_BAND_FIRMWARE_NAME=$BDF_FILE_NAME_5_G
			channel_array_name="channel_array_5g"
            ;;
        *)
            print_error "Invalid mode: $1. Use '2G' or '5G'"
            print_usage
            ;;
    esac

    for key in "${!channel_array_name[@]}"; do
        FREQ_BAND_CHANNEL_ARRAY["$key"]="${channel_array_name[$key]}"
    done
}

# -------------------------------------------------------
# Argument parser
# -------------------------------------------------------
parse_options() {
    while getopts "c:b:h" opt; do
        case $opt in
            c)
                FREQ_BAND_CHANNEL=$OPTARG
                ;;
            b)
                setup_freq_band "$OPTARG"
                ;;
            h)
                print_usage
			    exit 1
                ;;
            *)
                print_usage
			    exit 1
                ;;
        esac
    done

    # Check that both -c (channel) and -b (band) options were provided
    if [ -z "$FREQ_BAND" ]; then
        print_error "Option -b (band) is required."
		print_usage
	    exit 1
    fi

	if [ -z "$FREQ_BAND_CHANNEL" ]; then
        print_error "Option -c (channel) is required."
		print_usage
    	print_channels
		exit 1
    fi

	# Validate channel is valid for selected frequency band
	if [[ -z "${FREQ_BAND_CHANNEL_ARRAY[$FREQ_BAND_CHANNEL]}" ]]; then
		print_error "Invalid channel: $FREQ_BAND_CHANNEL for $FREQ_BAND band"
		print_usage
		print_channels
	    exit 1
	fi
}

# -------------------------------------------------------
# Check if a Qualcomm device is present
# -------------------------------------------------------
check_qcom_device() {
	if lspci | grep -q "Qualcomm Device"; then
		print_info "Qualcomm PCI Device found"
	else
		print_error "Qualcomm PCI Device not found"
		exit 1
	fi
}

# -------------------------------------------------------
# Check if the correct 
# -------------------------------------------------------
is_freq_band_fw_loaded() {
    # Compare the firmware files
    cmp -s "${FIRMWARE_PATH}${BDF_FILE_NAME}" "${FIRMWARE_PATH}${FREQ_BAND_FIRMWARE_NAME}"

    return $?
}

# -------------------------------------------------------
# Load kernel module
# -------------------------------------------------------
load_kernel_module() {
    force_reload=$1
    print_info "In load_kernel_module"

    if [ "$force_reload" -eq 1 ] 2>/dev/null; then
        print_info "Force reloading kernel module"
        # Unload the Wi-Fi module
        modprobe -r wifi_3_0 || {
            print_error "Failed to unload wifi_3_0 module"
            exit 1
        }
    else
        # check if module loaded
        lsmod | grep -q "wifi_3_0" && return 0 || echo "wifi_3_0 needs to be loaded"
    fi

    # Reload the Wi-Fi module
    MODULE_PATH="/extra" modprobe wifi_3_0 || {
        print_error "Failed to load wifi_3_0 module."
        exit 1
    }
    print_info "Kernel modules loaded successfully."
}

# -------------------------------------------------------
# Load correct firmware for frequency band
# -------------------------------------------------------
swap_band_firmware() {
    print_info "Swapping firmware binary for $FREQ_BAND: $FREQ_BAND_FIRMWARE_NAME"

    # Remount filesystem as read-write
    mount -o remount,rw /firmware || {
        print_error "Failed to remount /firmware as read-write."
        exit 1
    }

    # Copy the correct firmware file
    cp "${FIRMWARE_PATH}${FREQ_BAND_FIRMWARE_NAME}" "${FIRMWARE_PATH}${BDF_FILE_NAME}" || {
        print_error "Failed to copy firmware file."
        mount -o remount,ro /firmware
        exit 1
    }
    print_info "Firmware binary swapped successfully."

    # Remount as read-only
    mount -o remount,ro /firmware || {
        print_error "Failed to remount /firmware as read-only."
        exit 1
    }
}

# -------------------------------------------------------
# Setup wifi module
# -------------------------------------------------------
setup_wifi_module() {
	if is_freq_band_fw_loaded; then
		print_info "Correct firmware binary is loaded for $FREQ_BAND: $FREQ_BAND_FIRMWARE_NAME"
		load_kernel_module 0
	else
        swap_band_firmware
        load_kernel_module 1
    fi
}

# -------------------------------------------------------
# Create ath0 interface and configure band/channel
# -------------------------------------------------------
configure_interface() {
	cfg80211tool wifi0 enable_ol_stats 1
	wlanconfig ath0 create wlandev wifi0 wlanmode specialvap -cfg80211
	iw dev wifi0 interface add ath0 type __ap &>/dev/null
	
	# Configure ath0 for the selected band and channel
	cfg80211tool ath0 mode "$FREQ_BAND_MODE"
	cfg80211tool ath0 channel "$FREQ_BAND_CHANNEL" "$FREQ_BAND_ID"
	cfg80211tool ath0 ssid Openwrt_specialvap
	cfg80211tool ath0 hide_ssid 0
	cfg80211tool ath0 wds 0
	cfg80211tool ath0 backhaul 0
	cfg80211tool ath0 en_6g_sec_comp 1
	cfg80211tool ath0 uapsd 1
	cfg80211tool ath0 stafwd 0
	cfg80211tool ath0 ap_bridge 1

	ifconfig ath0 up

	# Show logs
	dmesg | grep "send_vdev_start_cmd_tlv" | tail -n 1
	dmesg | grep "htt_h2t_full_mon_cfg" | tail -n 1
}

# -------------------------------------------------------
# Main
# -------------------------------------------------------
parse_options "$@"
check_qcom_device
setup_wifi_module
configure_interface